/* OV518 Decompression Filter
 *
 * Copyright (c) 1999-2002 Mark W. McClelland. All rights reserved.
 * Portions Copyright 1998-2000 OmniVision Technologies
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <png.h>
#include <sys/mman.h>
#include <fcntl.h>
#include "ov518lib.h"

#define WIDTH  320
#define HEIGHT 240

//#define DYNQUANT
#define NOUV

/********************************** PROTOTYPES *****************************/
extern int decompressOV518(int, int, void *, int, void *, int, int,
			   struct frame *);

/*********************************** GLOBALS *******************************/

unsigned char qt[64];

/********************************* CONSTANTS *******************************/

unsigned long ZZTable[] = {
	0x00, 0x02, 0x03, 0x09,
	0x01, 0x04, 0x08, 0x0a,
	0x05, 0x07, 0x0b, 0x11,
	0x06, 0x0c, 0x10, 0x12,
	0x0d, 0x0f, 0x13, 0x19,
	0x0e, 0x14, 0x18, 0x1a,
	0x15, 0x17, 0x1b, 0x1e,
	0x16, 0x1c, 0x1d, 0x1f
};

/**************************************************************************/

void fatal_err(char *msg)
{
        fprintf(stderr, msg);
        exit(1);
}

/* From vidcat.c (License: GPL) */
void
put_image_png (FILE *out, char *image, int width, int height)
{
	int y;
	char *p;
	png_infop info_ptr;
	png_structp png_ptr = png_create_write_struct (PNG_LIBPNG_VER_STRING,
						NULL, NULL, NULL);
	png_bytep row_pointers[height];
        
	if (!png_ptr)
		return;
	info_ptr = png_create_info_struct (png_ptr);
	if (!info_ptr)
		return;

	png_init_io (png_ptr, out);
        
	png_set_IHDR (png_ptr, info_ptr, width, height,
					8, PNG_COLOR_TYPE_RGB, PNG_INTERLACE_NONE,
					PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);

	png_set_bgr (png_ptr);
	png_write_info (png_ptr, info_ptr);

	p = image;

	for (y = 0; y < height; y++) {
		row_pointers[y] = p;
		p += width * 3;
	}
        
	png_set_rows(png_ptr, info_ptr, row_pointers);
	png_write_png(png_ptr, info_ptr, 0, NULL);
	png_write_end (png_ptr, info_ptr);
}

/* LIMIT: convert a 16.16 fixed-point value to a byte, with clipping. */
#define LIMIT(x) ((x)>0xffffff?0xff: ((x)<=0xffff?0:((x)>>16)))

/* From ov511.c (License: GPL) */
static inline void
ov511_move_420_block(int yTL, int yTR, int yBL, int yBR, int u, int v, 
	int rowPixels, unsigned char * rgb)
{
#if 1
	const int rvScale = 91881;
	const int guScale = -22553;
	const int gvScale = -46801;
	const int buScale = 116129;
	const int yScale  = 65536;
	int r, g, b;

	g = guScale * u + gvScale * v;
	r = rvScale * v;
	b = buScale * u;

	yTL *= yScale; yTR *= yScale;
	yBL *= yScale; yBR *= yScale;

	/* Write out top two pixels */
	rgb[0] = LIMIT(b+yTL); rgb[1] = LIMIT(g+yTL); rgb[2] = LIMIT(r+yTL);
	rgb[3] = LIMIT(b+yTR); rgb[4] = LIMIT(g+yTR); rgb[5] = LIMIT(r+yTR);

	/* Skip down to next line to write out bottom two pixels */
	rgb += 3 * rowPixels;
	rgb[0] = LIMIT(b+yBL); rgb[1] = LIMIT(g+yBL); rgb[2] = LIMIT(r+yBL);
	rgb[3] = LIMIT(b+yBR); rgb[4] = LIMIT(g+yBR); rgb[5] = LIMIT(r+yBR);
#else
// GBR420
	/* Write out top two pixels */
	rgb[0] = u; rgb[1] = yTL; rgb[2] = v;
	rgb[3] = u; rgb[4] = yTR; rgb[5] = v;

	/* Skip down to next line to write out bottom two pixels */
	rgb += 3 * rowPixels;
	rgb[0] = u; rgb[1] = yBL; rgb[2] = v;
	rgb[3] = u; rgb[4] = yBR; rgb[5] = v;
#endif
}

/* Converts from planar YUV420 to RGB24. This is currently only
 * useful for postprocessing decompressed data, but will eventually replace
 * ov511_parse_yuv420_to_rgb().
 */
/* From ov511.c (License: GPL) */
static void 
yuv420p_to_rgb(int width, int height,
		     unsigned char *pIn0, unsigned char *pOut0, int bits)
{
	const int numpix = width * height;
	const int bytes = bits >> 3;
	int i, j, y00, y01, y10, y11, u, v;
	unsigned char *pY = pIn0;
	unsigned char *pU = pY + numpix;
	unsigned char *pV = pU + numpix / 4;
	unsigned char *pOut = pOut0;

	for (j = 0; j <= height - 2; j += 2) {
		for (i = 0; i <= width - 2; i += 2) {
			y00 = *pY;
			y01 = *(pY + 1);
			y10 = *(pY + width);
			y11 = *(pY + width + 1);

#ifdef DYNQUANT
			u = (*pU++);
			v = (*pV++);
#else
			u = (*pU++) - 128;
			v = (*pV++) - 128;
#endif

/* Disable UV */
#ifdef NOUV
			u = 0;
			v = 0;
#endif

//			u = (u * 577) >> 10;
//			v = (v * 730) >> 10;
//			u += 128;
//			v += 128;

//			u = (*pU++);
//			v = (*pV++);
//			u = 0;
//			v = 0;

			ov511_move_420_block(y00, y01, y10, y11, u, v,
					     width, pOut);
	
			pY += 2;
			pOut += 2 * bytes;

		}
		pY += width;
		pOut += width * bytes;

//		fprintf(stderr, "u=%d, v=%d ", u, v);
	}
}

int main(int argc, char **argv)
{
	int tmp;
//	int rc;
	char *inbuffer;
	char *outbuffer;
	int insize, outsize;
	FILE *in = stdin;
	FILE *out = stdout;
	struct frame f;
	int max_bufsize = WIDTH * HEIGHT * 4;
#ifdef DYNQUANT
	unsigned char qsxqt[] = {
		0x0f, 0x1f, 0x1f, 0x3f, 0x1f, 0x1f, 0x3f, 0x1f,
		0x3f, 0x7f, 0x7f, 0x3f, 0x1f, 0x3f, 0x7f, 0x3f,
		0x3f, 0x3f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
		0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0xff, 0xff, 0xff,        
		0x1f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x7f, 0x3f,
		0x3f, 0x7f, 0x7f, 0x3f, 0x7f, 0x7f, 0x7f, 0x7f,        
		0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
		0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0xff, 0xff, 0xff
	};
#else
	unsigned char qtY[] = OV518_YQUANTABLE;
	unsigned char qtUV[] = OV518_UVQUANTABLE;
#endif

	inbuffer = malloc(max_bufsize);
	if (! inbuffer)
		fatal_err("Couldn't malloc inbuffer\n");

	outbuffer = malloc(max_bufsize);
	if (! outbuffer)
		fatal_err("Couldn't malloc outbuffer\n");

	memset(inbuffer, 0, max_bufsize);
	memset(outbuffer, 0, max_bufsize);

// Read from file instead of stdin
#if 1
	in = fopen("out.raw", "r");
	if (! in) {
		perror("open");
		return 1;
	}
	fprintf(stderr, "Opened input file\n");
#endif

	tmp = fread(inbuffer, 1, max_bufsize, in);
	if (! tmp)
		fatal_err("Error reading from stdin\n");
	insize = tmp;

	fprintf(stderr, "Read %d bytes\n", insize);

	/* Set up the globals */
#ifdef DYNQUANT
	memcpy(qt, qsxqt, 64);
#else
	memcpy(qt, qtY, 32);
	memcpy(qt + 32, qtUV, 32);
#endif
	f.cinfo.rawLen = insize;
	f.cinfo.qt = qt;

	f.m_TempBufferLength = max_bufsize;
	f.m_pTempBuffer = malloc(max_bufsize);
	if (! f.m_pTempBuffer)
		fatal_err("Couldn't malloc tempbuffer\n");
	memset(f.m_pTempBuffer, 0, max_bufsize);

// Remember to subtract the QSxQT size from ActualRawFrameLength once it is
// passed within the frame
	outsize = decompressOV518(WIDTH,
				  HEIGHT,
				  outbuffer,
				  WIDTH * HEIGHT * 3 / 2, // outsize
				  inbuffer, 
				  max_bufsize,            // insize_max
				  insize,
				  &f);

// Decompressor will always return failure, since it doesn't know the real frame
// size.
#if 0
	if (! outsize)
		fatal_err("Error decompressing frame\n");

	fprintf(stderr, "Image is %d bytes...\n", outsize);
#endif

	yuv420p_to_rgb(WIDTH, HEIGHT, outbuffer, f.m_pTempBuffer, 24);
	put_image_png(out, f.m_pTempBuffer, WIDTH, HEIGHT);

	fprintf(stderr, "Wrote png file\n");

	free(f.m_pTempBuffer);
	free(inbuffer);
	free(outbuffer);

	return 0;
}
